@@ -37,8 +37,8 @@ void clara(int *n,  /* = number of objects */
 			 *	      1 : use R's RNG (and seed) */
 	   int *nrepr,
 	   int *nsel,
-	   int *nbest,/* x[nbest[j]] will be the j-th obs in the final sample */
-	   int *nr, int *nrx,
+	   int *nbest,/* x[nbest[j],] : the j-th obs in the final sample */
+	   int *nr, int *nrx,/* prov. and final "medoids" aka representatives */
 	   double *radus, double *ttd, double *ratt,
 	   double *ttbes, double *rdbes, double *rabes,
 	   int *mtt, double *obj,
@@ -61,14 +61,15 @@ void clara(int *n,  /* = number of objects */
 
     /* Local variables */
 
-    Rboolean nafs, kall, full_sample, lrg_sam, has_NA = *mdata;
-    int j, jk, jkk, js, jsm, jhalt, jran, l, n_sam;
-    int nsm, ntt, nad, nadv, kran, kans, nrun, nexap, nexbp;
-    int n_dys, nsamb, nunfs;
+    Rboolean nafs, kall, full_sample, lrg_sam, dyst_toomany_NA,
+	has_NA = *mdata;
+    int j, jk, jkk, js, jsm, jran, l, n_sam;
+    int nsm, ntt, nad, nadv, kran, kans, nrun, n_dys, nsamb, nunfs;
     double rnn, sky, zb, s, sx = -1., zba = -1.;/* Wall */
 
     *jstop = 0;
     rnn = (double) (*n);
+
     /* n_dys := size of distance array dys[] */
     n_dys = *nsam * (*nsam - 1) / 2 + 1;/* >= 1 */
     full_sample = (*n == *nsam);/* only one sub sample == full data */
@@ -78,22 +79,44 @@ void clara(int *n,  /* = number of objects */
 	n_sam = *n - *nsam;
     else
 	n_sam = *nsam;
-    nunfs = 0;
-    kall = FALSE;
 
-    if(*rng_R)
+    if(*trace_lev) Rprintf("C clara(): (nsam,n) = (%d,%d);%s\n", *nsam, *n,
+			   full_sample ? " 'full_sample',":
+			   (lrg_sam ? " 'large_sample',": ""));
+    if(*rng_R && !full_sample)
 	GetRNGstate();
     else /* << initialize `random seed' of the very simple randm() below */
 	nrun = 0;
 
+#define SET_kran_trace_print(_nr_)					\
+	kran= 1+ (int)(rnn* ((*rng_R)? unif_rand(): randm(&nrun)));	\
+	if (kran > *n) {/* should never happen */			\
+	    REprintf("** C clara(): random k=%d > n **\n", kran);	\
+	    kran = *n;							\
+	}								\
+	if(*trace_lev >= 4) {						\
+	    Rprintf("... {" #_nr_ "}");					\
+	    if(*rng_R) Rprintf("R unif_rand()");			\
+	    else       Rprintf("nrun=%5d", nrun);			\
+	    Rprintf(" -> k{ran}=%d\n", kran);				\
+	}
+
+
 /* __LOOP__ :  random subsamples are drawn and partitioned into kk clusters */
 
+    kall = FALSE; /* kall becomes TRUE iff we've found a "valid sample",
+		     i.e. one for which all d(j,k) can be computed */
+    nunfs = 0;
+    dyst_toomany_NA = FALSE;
     for (jran = 1; jran <= *nran; ++jran) {
-	jhalt = 0;
+	if(*trace_lev) Rprintf("C clara(): sample %d ", jran);
 	if (!full_sample) {/* `real' case: sample size < n */
 	    ntt = 0;
-	    if (kall/*was jran != 1 */ && nunfs != jran && !lrg_sam) {
-		/* nsel[] := sort(nrx[])   for the first j=1:k	?? */
+	    if (kall && nunfs+1 != jran && !lrg_sam) {
+		/* Have had (at least) one valid sample; use its representatives
+		 * nrx[] :  nsel[] := sort(nrx[])  for the first j=1:k */
+		if(*trace_lev >= 2) Rprintf(" if (kall && nunfs...): \n");
+
 		for (jk = 0; jk < *kk; ++jk)
 		    nsel[jk] = nrx[jk];
 		for (jk = 0; jk < *kk-1; ++jk) {
@@ -110,51 +133,47 @@ void clara(int *n,  /* = number of objects */
 		    nsel[jk] = nsm;
 		}
 		ntt = *kk;
+
 	    }
-	    else {
+	    else { /* no valid sample  _OR_  lrg_sam */
+		if(*trace_lev >= 2) Rprintf(" finding 1st... new k{ran}:\n");
+
 		/* Loop finding random index `kran' not yet in nrx[] : */
 	    L180:
-		kran = 1 + (int)(rnn * ((*rng_R)? unif_rand(): randm(&nrun)));
-		if(*trace_lev >= 3) {
-		    if(*rng_R)
-			Rprintf("... R unif_rand() -> k{ran}=%d\n", kran);
-		    else
-			Rprintf("... {180} nrun=%d -> k{ran}=%d\n", nrun,kran);
-		}
-		if (kall/*jran != 1*/) {
+		SET_kran_trace_print(180)
+
+		if (kall) {
 		    for (jk = 0; jk < *kk; ++jk)
 			if (kran == nrx[jk])
 			    goto L180;
 		}
 		/* end Loop */
+
 		nsel[ntt] = kran;
 		if (++ntt == n_sam)
 		    goto L295;
 	    }
 
 	    if(*trace_lev >= 2) {
-		Rprintf(".. kall(T/F)=%d , nsel[ntt=%d] = %d\n",
-			kall, ntt, nsel[ntt]);
-		if(*trace_lev >= 3) {
-		    Rprintf("... nrx[0:%d]= ",*kk-1);
-		    for (jk = 0; jk < *kk; jk++)
-			Rprintf("%d ",nrx[jk]); Rprintf("\n");
-		    Rprintf("... nsel[1:%d]= ",ntt);
-		    for (jk = 1; jk <= ntt; jk++)
-			Rprintf("%d ",nsel[jk]); Rprintf("\n");
+		Rprintf(".. kall: %s, ", (kall) ? "T" : "FALSE");
+		if(*trace_lev == 2) {
+		    Rprintf("nsel[ntt=%d] = %d\n", ntt, nsel[ntt]);
+		} else { /* trace_lev >= 3 */
+		    Rprintf("\n... nrx [0:%d]= ",*kk-1);
+		    for (jk = 0; jk < *kk; jk++) Rprintf("%d ",nrx[jk]);
+		    Rprintf("\n... nsel[0:%d]= ",ntt-1);
+		    for (jk = 0; jk < ntt; jk++) Rprintf("%d ",nsel[jk]);
+		    Rprintf("\n");
 		}
 	    }
 
 	    do {
-	    L210:
-		/* find `kran', a random `k' in {1:n},
+		/* Loop finding random index 'kran' in {1:n},
 		 * not in nrx[0:(k-1)] nor nsel[1:ntt] : */
-		kran = (int) (rnn * randm(&nrun) + 1.);
-		if(*trace_lev >= 3)
-		    Rprintf("... {210} nrun=%d -> k{ran}=%d\n", nrun,kran);
-		if (kran > *n) kran = *n;
+	    L210:
+		SET_kran_trace_print(210)
 
-		if (kall/*jran != 1*/ && lrg_sam) {
+		if (kall && lrg_sam) {
 		    for (jk = 0; jk < *kk; ++jk) {
 			if (kran == nrx[jk])
 			    goto L210;
@@ -181,31 +200,36 @@ void clara(int *n,  /* = number of objects */
 	    } while (ntt < n_sam);
 
 	L295:
+	    if(*trace_lev) Rprintf(" {295} [ntt=%d, nunfs=%d] ", ntt, nunfs);
 	    if (lrg_sam) {
 		/* have indices for smaller _nonsampled_ half; revert this: */
-		for (j = 1, nexap = 0, nexbp = 0; j <= *n; j++) {
-		    if (nsel[nexap] == j)
-			++nexap;
+		for (j = 1, jk = 0, js = 0; j <= *n; j++) {
+		    if (jk < n_sam && nsel[jk] == j)
+			++jk;
 		    else
-			nrepr[nexbp++] = j;
+			nrepr[js++] = j;
 		}
 		for (j = 0; j < *nsam; ++j)
 		    nsel[j] = nrepr[j];
+		if(*trace_lev >= 3) {
+		    Rprintf(".. nsel[1:%d]= ", *nsam);
+		    for (jk = 0; jk < *nsam; jk++) Rprintf("%d ",nsel[jk]);
+		}
 	    }
-	    if(*trace_lev)
-		Rprintf("C clara(): sample %d [ntt=%d, nunfs=%d] -> dysta2()\n",
-			jran, ntt, nunfs);
+	    if(*trace_lev) Rprintf(" -> dysta2()\n");
 	}
 	else { /* full_sample : *n = *nsam -- one sample is enough ! */
 	    for (j = 0; j < *nsam; ++j)
-		nsel[j] = j+1;/* <- still uses 1-indices for its *values*! */
+		nsel[j] = j+1;/* <- uses 1-indices for its *values*! */
 	}
 
 	dysta2(*nsam, *jpp, nsel, x, *n, dys, *diss_kind,
-	       jtmd, valmd, &jhalt);
-	if (jhalt == 1) {
+	       jtmd, valmd, &dyst_toomany_NA);
+	if(dyst_toomany_NA) {
 	    if(*trace_lev)
-		Rprintf("  dysta2() gave jhalt = 1 --> new sample\n");
+		Rprintf("  dysta2() gave dyst_toomany_NA --> new sample\n");
+	    dyst_toomany_NA = FALSE;
+	    ++nunfs;
 	    continue;/* random sample*/
 	}
 
@@ -218,7 +242,7 @@ void clara(int *n,  /* = number of objects */
 	} while (l+1 < n_dys);
 
 	if(*trace_lev >= 2)
-	    Rprintf(". clara(): s:= max dys[l=%d] = %g\n", l,s);
+	    Rprintf(". clara(): s:= max dys[l=%d] = %g;", l,s);
 
 	bswap2(*kk, *nsam, nrepr, dys, &sky, s,
 	       /* dysma */tmp1, /*dysmb*/tmp2,
@@ -228,11 +252,16 @@ void clara(int *n,  /* = number of objects */
 	      nrepr, nsel, dys, x, nr, &nafs, ttd, radus, ratt,
 	      ntmp1, ntmp2, ntmp3, ntmp4, ntmp5, ntmp6, tmp1, tmp2);
 
-	if (nafs) {
+	if (nafs) { /* couldn't assign some observation (to a cluster)
+		     * because of too many NA s */
 	    ++nunfs;
+	    if(*trace_lev >= 2)
+		Rprintf(" selec() -> 'NAfs'");
 	}
-	else if(!kall /*jran == 1*/ || zba > zb) {
-	    /* 1st proper sample  or  new best */
+	else if(!kall || zba > zb) { /* 1st proper sample  or  new best */
+	    kall = TRUE;
+	    if(*trace_lev >= 2)
+		Rprintf(" 1st proper or new best: zb= %g", zb);
 	    zba = zb;
 	    for (jk = 0; jk < *kk; ++jk) {
 		ttbes[jk] = ttd	 [jk];
@@ -243,37 +272,50 @@ void clara(int *n,  /* = number of objects */
 	    for (js = 0; js < *nsam; ++js)
 		nbest[js] = nsel[js];
 	    sx = s;
-	}
-
-	kall = TRUE;
+	} else if(*trace_lev >= 2) Rprintf(" zb= %g", zb);
+	if(*trace_lev >= 2) Rprintf("\n");
 
 	if(full_sample) break; /* out of resampling */
     }
 /* --- end random sampling loop */
-    if(*rng_R)
+    if(*rng_R && !full_sample)
 	PutRNGstate();
 
     if (nunfs >= *nran) { *jstop = 1; return; }
+    /* else */
+    if (!kall) { *jstop = 2; return; }
+
+    if(*trace_lev) {
+	Rprintf("C clara(): best sample _found_ ");
+	if(*trace_lev >= 2) {
+	    Rprintf("; nbest[1:%d] =\n c(", *nsam);
+	    for (js = 0; js < *nsam; ++js) {
+		Rprintf("%d", nbest[js]);
+		if(js+1 < *nsam) Rprintf(",");
+	    }
+	    Rprintf(")\n");
+	}
+	Rprintf(" --> dysta2(nbest), resul(), end\n");
+    }
 
 
 /*     for the best subsample, the objects of the entire data set
      are assigned to their clusters */
 
-    if (!kall) { *jstop = 2; return; }
-
-    if(*trace_lev) {
-	Rprintf("C clara(): sample _found_ --> dysta2(nbest), resul(), end\n");
-    }
     *obj = zba / rnn;
-    dysta2(*nsam, *jpp, nbest, x, *n, dys, *diss_kind, jtmd, valmd, &jhalt);
-
+    dysta2(*nsam, *jpp, nbest, x, *n, dys, *diss_kind, jtmd, valmd,
+	   &dyst_toomany_NA);
+    if(dyst_toomany_NA) {
+	REprintf(" *** SHOULD NOT HAPPEN: clara() -> dysta2(nbest) gave toomany_NA\n");
+	return;
+    }
     resul(*kk, *n, *jpp, *diss_kind, has_NA, jtmd, valmd, x, nrx, mtt);
 
     if (*kk > 1)
 	black(*kk, *jpp, *nsam, nbest, dys, sx, x,
 	      /* compute --> */
 	      avsyl, ttsyl, sylinf,
-	      ntmp1, ntmp2, ntmp3, ntmp4, tmp1, tmp2);
+	      ntmp1, ntmp2, ntmp3, ntmp4, /* syl[] */ tmp1, tmp2);
     return;
 } /* End clara() ---------------------------------------------------*/
 #undef tmp1
@@ -289,7 +331,7 @@ void clara(int *n,  /* = number of objects */
 
 void dysta2(int nsam, int jpp, int *nsel,
 	    double *x, int n, double *dys, int diss_kind,
-	    int *jtmd, double *valmd, int *jhalt)
+	    int *jtmd, double *valmd, Rboolean *toomany_NA)
 {
 /* Compute Dissimilarities for the selected sub-sample	---> dys[,] */
 
@@ -303,7 +345,7 @@ void dysta2(int nsam, int jpp, int *nsel,
 	lsel = nsel[l];
 	if(lsel <= 0 || lsel > n)
 	    REprintf(" ** dysta2(): nsel[l= %d] = %d is OUT\n", l, lsel);
-	for (k = 0; k < l; ++k) {
+	for (k = 0; k < l; ++k) { /* compute d(nsel[l], nsel[k]) {if possible}*/
 	    ksel = nsel[k];
 	    if(ksel <= 0 || ksel > n)
 		REprintf(" ** dysta2(): nsel[k= %d] = %d is OUT\n", k, ksel);
@@ -313,13 +355,10 @@ void dysta2(int nsam, int jpp, int *nsel,
 	    for (j = 0; j < jpp; ++j) {
 		lj = lsel-1 + j * n;
 		kj = ksel-1 + j * n;
-		if (jtmd[j] < 0) {
+		if (jtmd[j] < 0) { /* x[,j] has some Missing (NA) */
 		    /* in the following line (Fortran!), x[-2] ==> seg.fault
 		       {BDR to R-core, Sat, 3 Aug 2002} */
-		    if (x[lj] == valmd[j]) {
-			continue /* next j */;
-		    }
-		    if (x[kj] == valmd[j]) {
+		    if (x[lj] == valmd[j] || x[kj] == valmd[j]) {
 			continue /* next j */;
 		    }
 		}
@@ -329,15 +368,15 @@ void dysta2(int nsam, int jpp, int *nsel,
 		else
 		    clk += fabs(x[lj] - x[kj]);
 	    }
-	    if (npres == 0) {
-		*jhalt = 1;
+	    if (npres == 0) {/* cannot compute d(.,.) because of too many NA */
+		*toomany_NA = TRUE;
 		dys[nlk] = -1.;
 	    } else {
 		d1 = clk * (((double) (jpp)) / (double) npres);
 		dys[nlk] = (diss_kind == 1) ? sqrt(d1) : d1 ;
 	    }
-	}
-    }
+	} /* for( k ) */
+    } /* for( l ) */
     return;
 } /* End dysta2() -----------------------------------------------------------*/
 
@@ -353,13 +392,31 @@ double randm(int *nrun)
     return ((double) (*nrun) / 65536.);
 } /* randm() */
 
+int ind_2(int l, int j)
+{
+/* Utility, originally FORTRAN,  called "meet"; called from CLARA, PAM & TWINS.
+ * Original code had MEET(), MEET2(), and MEET3() in the 3 source files.
+
+ * ind_2(l,j) returns the *index* of dys() where diss. d(l,j) is stored:
+ *        d(l,j) == dys[ind_2(l,j)]
+ *
+ * MM: changed to work with 0-origin matrices  dys[], but l,j are >= 1
+ */
+    if(l > j)
+	return (l-2)*(l-1)/2 + j;
+    else if(l == j)
+	return 0;/* and the first element, dys[0] is := 0.  permanently! */
+    else /* l < j */
+	return (j-2)*(j-1)/2 + l;
+}
+
 /* bswap2() : called once [per random sample] from clara() : */
 void bswap2(int kk, int nsam, int *nrepr,
 	    double *dys, double *sky, double s,
 	    double *dysma, double *dysmb, double *beter)
 {
-    int j, ja, k, kbest = -1, nbest = -1;/* init for -Wall */
-    int nkj, njn, njaj, nny, nmax;
+    int j, ja, k, kbest = -1, nbest = -1, nmax = -1;/* init for -Wall */
+    int nkj, njn, njaj, nny;
 
     double ammax, small, asky, dzsky, dz, cmd;
 
@@ -367,15 +424,15 @@ void bswap2(int kk, int nsam, int *nrepr,
     --nrepr;
     --beter;
 
-    --dys;/*index via meet() */
     --dysma;	--dysmb;
 
+    s = s * 1.1 + 1.;/* value larger than all dissimilarities */
 
 /* ====== first algorithm: BUILD. ====== */
 
     for (j = 1; j <= nsam; ++j) {
 	nrepr[j] = 0;
-	dysma[j] = s * 1.1 + 1.;
+	dysma[j] = s;
     }
 
     for(nny = 0; nny < kk; nny++) {
@@ -383,8 +440,7 @@ void bswap2(int kk, int nsam, int *nrepr,
 	    if (nrepr[ja] == 0) {
 		beter[ja] = 0.;
 		for (j = 1; j <= nsam; ++j) {
-		    njaj = F77_CALL(meet)(&ja, &j);
-		    cmd = dysma[j] - dys[njaj];
+		    cmd = dysma[j] - dys[ ind_2(ja, j)];
 		    if (cmd > 0.)
 			beter[ja] += cmd;
 		}
@@ -399,7 +455,7 @@ void bswap2(int kk, int nsam, int *nrepr,
 	}
 	nrepr[nmax] = 1;
 	for (j = 1; j <= nsam; ++j) {
-	    njn = F77_CALL(meet)(&nmax, &j);
+	    njn = ind_2(nmax, j);
 	    if (dysma[j] > dys[njn])
 		dysma[j] = dys[njn];
 	}
@@ -420,11 +476,11 @@ void bswap2(int kk, int nsam, int *nrepr,
 L60:
 
     for (j = 1; j <= nsam; ++j) {
-	dysma[j] = s * 1.1 + 1.;
-	dysmb[j] = s * 1.1 + 1.;
+	dysma[j] = s;
+	dysmb[j] = s;
 	for (ja = 1; ja <= nsam; ++ja) {
 	    if (nrepr[ja] != 0) {
-		njaj = F77_CALL(meet)(&ja, &j);
+		njaj = ind_2(ja, j);
 		if (dys[njaj] < dysma[j]) {
 		    dysmb[j] = dysma[j];
 		    dysma[j] = dys[njaj];
@@ -441,8 +497,8 @@ L60:
 		if (nrepr[ja] != 0) {
 		    dz = 0.;
 		    for (j = 1; j <= nsam; ++j) {
-			njaj = F77_CALL(meet)(&ja, &j);
-			nkj  = F77_CALL(meet)(&k, &j);
+			njaj = ind_2(ja, j);
+			nkj  = ind_2(k, j);
 			if (dys[njaj] == dysma[j]) {
 			    small = dysmb[j];
 			    if (small > dys[njaj])
@@ -496,14 +552,10 @@ void selec(int kk, int n, int jpp, int diss_kind,
     --rdnew; --ttnew; --npnew; --nrnew; --nsnew;
     --new;
 
-    --dys;
-
-
-    /* nafs := TRUE  if a distance cannot be calculated*/
+    /* nafs := TRUE  if a distance cannot be calculated (because of NA s)*/
     *nafs = FALSE;
 
-/* identification of representative objects, and initializations */
-
+    /* identification of representative objects, and initializations */
     jk = 0;
     for (j = 1; j <= nsam; ++j) {
 	if (nrepr[j] != 0) {
@@ -583,7 +635,7 @@ void selec(int kk, int n, int jpp, int diss_kind,
 	    if (!pres) { /* found nothing */
 		*nafs = TRUE; return;
 	    }
-	} /* if (has_NA..) else */
+	} /* else: has_NA */
 
 	if (diss_kind == 1)
 	    dnull = sqrt(dnull);
@@ -645,17 +697,15 @@ void selec(int kk, int n, int jpp, int diss_kind,
 		    continue /* next kb */;
 
 		npb = np[kb];
-		npab = F77_CALL(meet)(&npa, &npb);
+		npab = ind_2(npa, npb);
 		if (nstrt == 0)
 		    nstrt = 1;
 		else if (dys[npab] >= ratt[ka])
 		    continue /* next kb */;
 
 		ratt[ka] = dys[npab];
-		if (ratt[ka] != 0.)
-		    continue /* next kb */;
-
-		ratt[ka] = -1.;
+		if (ratt[ka] == 0.)
+		    ratt[ka] = -1.;
 	    }
 	    if (ratt[ka] > -0.5)
 		ratt[ka] = radus[ka] / ratt[ka];
@@ -759,6 +809,7 @@ void black(int kk, int jpp, int nsam, int *nbest,
 	   double *dys, double s, double *x,
 	   /* --> Output : */
 	   double *avsyl, double *ttsyl, double *sylinf,
+	   /* but the following output vectors are never by clara() : */
 	   int *ncluv, int *nsend, int *nelem, int *negbr,
 	   double *syl, double *srank)
 {
@@ -776,13 +827,9 @@ void black(int kk, int jpp, int nsam, int *nbest,
 /* Parameter adjustments */
     --avsyl;
 
-    --srank;
-    --syl;
-    --negbr;
-    --nelem;
-    --nsend;
+    --srank;    --syl;
+    --negbr; --nelem; --nsend;
     --ncluv;	--nbest;
-    --dys;
 
     sylinf_dim1 = nsam;
     sylinf_offset = 1 + sylinf_dim1 * 1;
@@ -823,7 +870,7 @@ void black(int kk, int jpp, int nsam, int *nbest,
 		    for (l = 1; l <= nsam; ++l) {
 			if (ncluv[l] == nclu) {
 			    ++nbb;
-			    db += dys[F77_CALL(meet)(&nj, &l)];
+			    db += dys[ind_2(nj, l)];
 			}
 		    }
 		    btt = (double) nbb;
@@ -841,7 +888,7 @@ void black(int kk, int jpp, int nsam, int *nbest,
 	    dysa = 0.;
 	    for (l = 1; l <= ntt; ++l) {
 		nl = nelem[l];
-		dysa += dys[F77_CALL(meet)(&nj, &nl)];
+		dysa += dys[ind_2(nj, nl)];
 	    }
 	    att = (double) (ntt - 1);
 	    dysa /= att;
